;******************************************************************************
; What is it for?:
; ***************************
; Character receive and display for 2/4 line x 8/16/20/40 char LCD.
;            --> 40x4 Char. LCD is NOT supported <--
;
; The basic task of this Interface is to unburden the master to
; initialize,generate timing-signals and control the LCD.
;
;------------------------------------------------------
; Basic Idea found at Dick Cappels' project pages at
; http://www.projects.cappels.org
;------------------------------------------------------
;
; This Interface accept the following Control-Codes for the LCD
;
; Clearscreen		    Code $01
; Cursor Home			    Code $02
; LCD On			Code $03
; LCD Off		    Code $04
; Cursor On		    Code $05
; Cursor Off		Code $06
; Blink On		    Code $07
; Blink Off		    Code $08
; Shift Cursor Left	    Code $09
; Shift Cursor Right	    Code $0A
; Shift Display Left	    Code $0B
; Shift Display Right	    Code $0C
; Carriage Return	Code $0D
; Backspace		    Code $0E
; Line Two		    Code $0F
; Line Three		    Code $10
; Line Four		    Code $11
; Clear Line		Code $12
; Backlight On		Code $13
; Backlight Off		Code $14
; 
; Characters are transmitted from a Master to this hardware by a
; TTL-compatible serial RS232 interface.This interface is receive only
; and will not transmit anything back to the Master.
;
; Requires ATTiny2313 or equivalent
; A Baudrate-Crystal with 3.6864 MHz was choosen to eliminate possible
; RS232 timing problems. 
;
; The display is driven in 4 bit mode. My LCD display uses a Samsung
; KS0070B controller that appears to differ from the Sharp or Hitachi HD44780
; controller in that way,that the Samsung requires an additional 4 bit write
; operation during "Function set" when in 4 bit mode. This was accomplished
; by writing the "Function set" command twice to the controller instead of once
; I suspect that this would work without modification (?) for the Sharp/Hitachi
; controller, as the second write would be a redundant command.
;
;             --> See comments in subroutine "InitDisplay" <-- 
;
;
; --> Please DOUBLE-check the pinout of your LCD display.I have seen <--
; --> different pin-layouts for "standard" 16x2 LCD-Displays !!!!    <--
;
; 
; ATTiny2313 connections:
;
; Pin		Function (Connection)
; ---------------------------------
; RXD(PD0)	TTL-RS232 Serial In from Master(logic 1= no receive-data)
; PD1		R/S pin on LCD (pin 4)
; PD2		E pin on LCD (pin 6)
; PD3-PD6	D4-D7 on LCD (pins 11-14)
;			D0-D3 on LCD (pins 7-10) -> ground
;           R/W on LCD (pin 5)       -> ground
; PB0       Backlight control via Transistor to LCD (pin15)
;
; Set Fuses to : 
; CKSEL 3..0 = 1101,SUT 1..0 = 11 (Crystal Oszillator,slowly rising Power)
; BODLEVEL = 100 (Brown-Out at 4,3Volt)		
; 
; The receive data rate is set to 9600/1/N.
; 
; Code description:
;************************ 
; A 64 byte interrupt controlled circular buffer is used to capture
; and store incoming data.
; This buffer is needed to store incoming data during some specific LCD-Commands.
; For example : if the LCD receives a "Clear-Screen" command, the LCD is busy
; for approx. 2msec.During this 2msec it can happen, that the Master sends other
; commands/characters to the LCD which will be lost, when not buffered.
; (2msec at 9600 Baud equals approx. 2 Characters received from the Master)
; ----------------------------------------------------------------------------------
; After Power-Up the Master MUST wait for about 50 msec to allow the LCD to be
; initialized.During this 50 msec the LCD is NOT able to respond to any
; characters/control-codes sent by the Master.  
; ----------------------------------------------------------------------------------
; The main loop of the program named "Wait_Char" continuously checks to see
; if new characters have been written to the buffer and processes new characters.
; Displayable characters, defined as those with ASCII codes above $1F are
; shown on the display, codes <= $1F are defined as Control-Codes for the Display.
;
; ----------------------------------------------------------------------------------
; Adapt Constants "LineLenght" and "NoOfLines" to your LCD-Display
; 
;*********************************************************************************** 
.nolist
.include "tn2313def.inc"
.list

.def	inchar		=r1		;Char received by UART
.def    temp    	=r16    ;Temporary register
.def	temp1		=r17	;Temporary register
.def    temp2  		=r18	;Temporary register.
.def    LCD_XPos	=r19	;Row (0,1,2..14,15)
.def    LCD_Ypos	=r20	;Line (1,2)

;	XL			;UART circular buffer write ponter.
;	YL			;UART circular buffer read pointer

   
;-------------------Define Constants ----------------------------------------   
;Assign I/O pins.
.equ OE			=4		;PD2 display (E) enable
.equ RS			=2		;PD1 display (RS) register select
.equ LCD_Port 	= PORTD
.equ LCD_DDR	= DDRD
.equ BL_Port 	= PORTB
.equ BL_DDR		= DDRB
.equ BL			= 0		;Backlight on PB0
;----------------------------------------------------------------------------
;Allocate buffer in Stack:
;Circular buffer (for incoming characters) from $60 through $A0.
;This buffer will wrap around when full (after storing a character
;in location $A0 the next incoming character will be stored at location $60)
; 
.equ	circbufsiz	=64
.equ	cbufbot		=$60					;Bottom of UART receive buffer.
.equ	cbuftop		=cbufbot + circbufsiz   ;Top of UART receive buffer ($A0).

;---------- LCD Constants  --> 40x4 Char. LCD is NOT supported <--  ---------
.equ	LineLenght	=16				;change for 8/16/20/40 Chars in 1 Line
.equ    NoOfLines	=2				;change for 2/4 Line Display
.equ	FirstLine_Add 	=0b10000000 ;Base address of first LCD line
.equ	SecondLine_Add	=0b11000000 ;Base address of second LCD line
.equ	ThirdLine_Add 	=0b10010100	;Base address of third LCD line
.equ	FourthLine_Add 	=0b11010100	;Base address of fourth LCD line
;---------- Control Code Constants ------------------------------------------
.equ	Code_min = 1		;minimum value of control-code
.equ	Code_max= $14		;maximum value of control-code
;---------- Set Baudrate ----------------------------------------------------
.equ UBRR_VAL   = 23		;9600 Baud at 3.6864 MHz Crystal

;*************  End of Definitions *****************

.CSEG
.ORG 0x0000			

	rjmp Reset 	;Reset Handler
	reti		;External Int. 0 Handler
	reti		;External Int. 1 Handler
	reti		;Timer1 Capture Handler
	reti		;Timer1 Compare A Handler
	reti		;Timer1 Overflow Handler
	reti		;Timer0 Overflow Handler
	rjmp UartRecInt		;USART RX complete Handler
	reti		;USART UDR empty Handler
	reti		;USART TX complete Handler
	reti		;Analog Comparator Handler
	reti		;Pin change Interrupt
	reti		;Timer1 Compare B Handler
	reti		;Timer0 Compare A Handler
	reti		;Timer0 Compare B Handler
	reti		;USI Start Handler
	reti		;USI Overflow Handler
	reti		;EEPROM Ready Handler
	reti		;Watchdog Overflow Handler
	   
;*****************************************************************
;***       Init-Routine (Starting at Label "reset"             ***
;***   and ending at Label "Wait_Char") needs about 50 msec    ***
;***   During this time (at Power-Up) no character(s) should   ***
;***  be sent to the LCD because the character(s) is/are lost! ***
;*****************************************************************
Reset:
	ldi temp,RAMEND		;init Stack
	out spl,temp
	ldi temp,0b01111110 ;PORTD = all outputs except bit 0   
    out LCD_DDR,temp		
	ldi temp,0b01111110    
    out LCD_Port,temp
	sbi	BL_DDR,BL		;PB0 (Backlight) as Output
	rcall	InitDisplay	;Initialize LCD display module
	ldi	temp,UBRR_VAL	;baudconstant	
	out	ubrrl,temp		;define baudrate
	ldi	temp,(1<<UCSZ1)|(1<<UCSZ0)	;8 Bit Data,1 Stop,No Parity
	out	UCSRC,temp
	sbi	ucsrb,rxen		;Enable the receiver..
	ldi	XL,cbufbot		;Set X and Y Character-Buffer pointers to bottom
	ldi	YL,cbufbot		;of circular Buffer
    sbi	ucsrb,RXCIE		;Enable UART receive Interrupt
	sbi	BL_Port,BL		;switch Backlight on
    sei					;Enable global Interrupt
	rjmp Clscreen		;Clear Screen,set LCD_Xpos to Row0,LCD_Ypos to Line1

 Wait_Char:				;Waiting for new data in circular buffer
	cp	XL,YL			;Is circular buffer read/write pointer equal?
	breq	empty		;If so, there is no new data in the buffer.
	ld	temp,Y+			;If pointers are not equal copy next char 
						;from circular-buffer to temp.
	cpi	YL,cbuftop + 1	;If top of circular-buffer reached,
						;reset pointer to bottom of buffer.	
	brne	NoZeroYL		
	ldi	YL,cbufbot		
NoZeroYL:
		;Handle the new character as either a control char or a displayable char.
		;If temp is a control char,test for LCD-Commands, otherwise ignore char
	cpi	temp,$20		;If not a control char branch to displayable char routine..
	brsh Display_Char	;Displayable Chars : 0x20 to 0xFF

;-------- Char is control character ---------

	subi    temp,Code_min             	;set offset for pointer to CodeTable
    cpi     temp,(Code_max-Code_min+1)	;check, that offset is between MIN and MAX
    brsh    No_LCD_Control             	;offset not in range:no known Control-Char
    ldi     ZL,low(CodeTable)			;load pointer
    ldi     ZH,high(CodeTable)
    add     ZL,temp						;add offset to pointer
    ldi     temp,0                   
    adc     ZH,temp
    ijmp								;jump in CodeTable
 
No_LCD_Control:
Empty:
	rjmp	Wait_Char		;look for next incoming char

CodeTable:	;perform LCD-Action according to received Control-Code
    rjmp    Clscreen		;Code $01
    rjmp    Home			;Code $02
    rjmp    LCD_On			;Code $03
    rjmp    LCD_Off			;Code $04
    rjmp    Cursor_On		;Code $05
    rjmp    Cursor_Off		;Code $06
    rjmp    Blink_On		;Code $07
    rjmp    Blink_Off		;Code $08
    rjmp    ShiftCursorL	;Code $09
    rjmp    ShiftCursorR	;Code $0A
    rjmp    ShiftDisplayL	;Code $0B
    rjmp    ShiftDisplayR	;Code $0C
    rjmp    CarriageReturn	;Code $0D
    rjmp    BackSpace		;Code $0E
    rjmp    LineTwo			;Code $0F
    rjmp    LineThree		;Code $10
    rjmp    LineFour		;Code $11
    rjmp    Clear_Line		;Code $12
	rjmp	BL_On			;Code $13
	rjmp	BL_Off			;Code $14

;--------  Char is displayable  ---------
Display_Char:
	rjmp ItsDisplayable	

;******** perform LCD-Commands *********		
Clscreen:					;Clear Screen (Code $01)
    ldi temp,0b00000001	
    rcall SendCommand 
    rcall delay2ms
	rcall delay2ms
    ldi LCD_Xpos,0
    ldi LCD_Ypos,1
	rjmp	Wait_Char

Clear_Line:					;Clears line where cursor is on (Code $12)
	ldi		temp,0
	mov		temp2,LCD_Ypos
	rcall	LCD_GotoXY
	ldi temp2,LineLenght
Clear_Loop:
	ldi	temp,$20
	rcall SendData
	dec	temp2
	brne	Clear_Loop
	ldi		LCD_Xpos,0
	mov		temp,LCD_Xpos
	mov		temp2,LCD_Ypos
	rcall	LCD_GotoXY
	rjmp	Wait_Char			

LCD_On:							;Display on (Code $03)
	ldi temp,0b00001110
	rcall SendCommand 
	rjmp	Wait_Char

LCD_Off:					;Display off (Code $04)
	ldi temp,0b00001000
	rcall SendCommand 
	rjmp	Wait_Char

Cursor_On:				;Cursor On (Code $05)
	ldi temp,0b00001110
	rcall SendCommand 
	rjmp	Wait_Char

Cursor_Off:					;Cursor Off (Code $06)
	ldi temp,0b00001100
	rcall SendCommand 
	rjmp	Wait_Char

Blink_On:					;Cursor blink On (Code $07)
	ldi temp,0b00001111
	rcall SendCommand 
	rjmp	Wait_Char

Blink_Off:					;Cursor blink Off (Code $08)
	ldi temp,0b00001110
	rcall SendCommand 
	rjmp	Wait_Char

ShiftCursorL:				;shift Cursor left (Code $09)
	ldi temp,0b00010000
	rcall SendCommand
	
	dec LCD_Xpos
	 
	rcall delay50us
	rjmp	Wait_Char

ShiftCursorR:				;shift Cursor right (Code $0A)
	ldi temp,0b00010100
	rcall SendCommand
	
	inc LCD_Xpos
	 
	rjmp	Wait_Char

ShiftDisplayL:				;shift Display Left (Code $0B)
	ldi temp,0b00011000
	rcall SendCommand 
	rjmp	Wait_Char

ShiftDisplayR:				;shift Display Right (Code $0C)
	ldi temp,0b00011100
	rcall SendCommand 
	rjmp	Wait_Char

CarriageReturn:				;Carriage Return (Code $0D)
							;moves cursor to next lower line, pos.0
							;if not allready in last (lowest) line.
	cpi LCD_Ypos,NoOfLines	;If in last (lowest) line,cursor position
	brlo	Do_CR			;will not be changed
	rjmp End_Chk_CarriageReturn
Do_CR:
	ldi		LCD_Xpos,0
	mov		temp,LCD_Xpos
	inc		LCD_Ypos
	mov		temp2,LCD_Ypos
	rcall	LCD_GotoXY	
End_Chk_CarriageReturn:
	rjmp	Wait_Char

BackSpace:					;BackSpace (Code $E)
	cpi LCD_Xpos,0			;allready at Row 0?
	brne Dec_Xpos			;no
	ldi temp,$20 			;erases the char. left of Curser-Position
	rcall SendData			;but stops at current line Pos. 0
	dec LCD_XPos
	mov temp,LCD_Xpos
	mov temp2,LCD_Ypos
	rcall LCD_GotoXY 
	rjmp End_Backspace
Dec_Xpos:
	dec Lcd_Xpos
	mov temp,LCD_Xpos
	mov temp2,LCD_Ypos
	rcall LCD_GotoXY
	ldi temp,$20
	rcall SendData
	dec Lcd_Xpos
	mov temp,LCD_Xpos
	mov temp2,LCD_Ypos
	rcall LCD_GotoXY
End_Backspace:
	rjmp	Wait_Char

Home:						;Cursor to start of line 1
	ldi temp,0b00000010
	rcall SendCommand 
    rcall delay2ms
	rcall delay2ms
    ldi LCD_Xpos,0			;indicate that we are row 0
    ldi LCD_Ypos,1			;indicate that we are on line 1
	rjmp	Wait_Char

LineTwo:					;Cursor to start of line 2
	ldi     temp,SecondLine_Add
	rcall SendCommand
	ldi	LCD_XPos,0			;indicate that we are row 0
	ldi	LCD_YPos,2			;indicate that we are on line 2
	rjmp	Wait_Char

LineThree:					;Cursor to start of line 3
	ldi     temp,ThirdLine_Add
	rcall SendCommand
	ldi	LCD_XPos,0			;indicate that we are row 0
	ldi	LCD_YPos,3			;indicate that we are on line 3
	rjmp	Wait_Char

LineFour:					;Cursor to start of line 4
	ldi     temp,FourthLine_Add
	rcall SendCommand
	ldi	LCD_XPos,0			;indicate that we are row 0
	ldi	LCD_YPos,4			;indicate that we are on line 4
	rjmp	Wait_Char

BL_On:						;BackLight On (Code $13)
	sbi	BL_Port,BL
	rjmp	Wait_Char

BL_Off:						;BackLight Off (Code $14)
	cbi	BL_Port,BL
	rjmp	Wait_Char

;-----------------------------------------
ItsDisplayable:	;If not a control char then write to display
	rcall	SendData
	rjmp	Wait_Char	

;-------------------------------------------------------------------     
;Uart interrupt service.Store received char into the circular buffer
;-------------------------------------------------------------------
UartRecInt:
	push	temp
	in	temp,sreg
	push	temp
	in	inchar,udr		;Get incoming char from UART.
	st	X+,inchar		;store in circular buffer and increment pointer		
	cpi	XL,cbuftop + 1	;top of circular buffer reached?
	brne	NoZeroXL
	ldi	XL,cbufbot		;yes,buffer-overflow.start at bottom again,
						;old data will be overwritten
NoZeroXL:
	pop	temp
	out	sreg,temp
	pop	temp
	reti

;*******************************************
;LCD_GOTOXY
; temp contains the column (x=0,1...38,39)
; temp2 contains the Line (y=1,2,3,4)
;*******************************************
LCD_GotoXY:
	ldi temp1,FirstLine_Add
	cpi temp2,1          ;Check to see which line
	breq Calc_Xpos
	ldi temp1,SecondLine_Add
	cpi	temp2,2
	breq Calc_Xpos
	ldi temp1,ThirdLine_Add
	cpi	temp2,3
	breq Calc_Xpos
	ldi temp1,FourthLine_Add
Calc_Xpos:
	push temp
	or temp,temp1
	rcall SendCommand    ;Set DDRAM Address to LCD
	pop temp
	mov LCD_Xpos,temp    ;update the cursor x and y position
	mov LCD_Ypos,temp2
	ret

SendCommand:
	push    temp		;Save a copy of the data
	andi    temp,$F0	;Mask off lower nibble.
	lsr		temp		;--->Port D has 7 Bits only<----
	sbr     temp,OE    	;set E bit high.
	out     LCD_Port,temp	;Write upper nibble to display.
	rcall	delay2us	;Wait 2 us.			
	pop		temp		;Retrieve copy of input data.
	push	temp
	cbi     LCD_Port,2     ;Set E low to clock in data.
	swap    temp 		;Swap upper and lower nibbles.
	andi    temp,$F0	;Mask off lower nibble.
	lsr		temp		;--->Port D has 7 Bits only<----
	sbr     temp,OE		;set E high.
	out     LCD_Port,temp	;Write lower nibble to LCD.
	rcall	delay2us	;Wait 2 us.		       
	cbi     LCD_Port,2     ;set OE low to clock in data.
	rcall	Delay50us	;Wait about 50 us before next action
	pop		temp
	ret    				        	

SendData:
	push    temp		;Save a copy of the data
	inc 	LCD_XPos
	cpi 	LCD_XPos,LineLenght+1	;end of line reached?
	brlo 	cont					;no
	pop 	temp					;yes,inhibit output to LCD if
	dec 	LCD_XPos				;at end of line
	rjmp 	End_SendData
cont:
	andi    temp,$F0	;Mask off lower nibble.
	lsr		temp		;--->Port D has 7 Bits only<----
	sbr     temp,OE    	;set E bit high.
	sbr     temp,RS		;Set data/command bit high ->write data
	out     LCD_Port,temp	;Write upper nibble to display.
	rcall	delay2us	;Wait 2 us.			
	pop		temp		;Retrieve copy of input data.
	cbi     LCD_Port,2     ;set E low to clock in data.
	swap    temp 		;Swap upper and lower nibbles.
	andi    temp,$F0	;Mask off lower nibble
	lsr		temp		;--->Port D has 7 Bits only<----
	sbr     temp,OE		;Make E high.
	sbr     temp,RS		;Set data/command bit high ->write data
	out     LCD_Port,temp	;Write lower nibble to LCD.
	rcall	delay2us	;Wait 2 us.		       
	cbi     LCD_Port,2     ;set OE low to clock in data.
	rcall	Delay50us	;Wait about 50 us before next action
End_SendData:
	ret        


InitDisplay:			;Initialize Display to 4Bit-Mode
;------ Wait 15ms or more ---------
	ldi	temp,10
ID1:
	rcall	delay2ms			
	dec	temp
	brne	ID1
;------ First Function set ------
	ldi     temp,0b00111000		;Function set+E=high
	out     LCD_Port,temp

;------ Wait 4 ms or more ------
	rcall	delay2ms			
	rcall	delay2ms
	rcall	delay2ms
;------ Second Function set ------	
	ldi     temp,0b00111000		;Function set+E=high
	out     LCD_Port,temp
;------ Wait 100usec or more ------
	rcall	delay2ms
;------ Third Function set ------			
	ldi     temp,0b00111000     ;Function set+E=high
	out     LCD_Port,temp
;------ ;Function set 4 bit mode, 2 lines 5X7 pixels ------
	ldi  	temp,0b00101000 	;upper nibble sets 4 bit mode,
								;lower nibble:N=1(dual line display),F=0(5x7 Dot)
	rcall   SendCommand
;***************** for KS0066 or equivalent only ***************			
	ldi     temp,0b00101000 	;delete FOR SHARP/HITACHI LCD-Controller
	rcall   SendCommand			;delete FOR SHARP/HITACHI LCD-Controller
;***************************************************************
;------ 	
	ldi     temp,0b00001000		;Display, cursor and blink off.
								;D=0(Display off),C=0(Cursor off),B=0(no blink)                                    	
	rcall   SendCommand
;------
	ldi     temp,0b00000001     ;Clear display
	rcall   SendCommand
	rcall   delay2ms     		;Need to wait more than 1.6 msec after clear.
	rcall   delay2ms 
;------
	ldi     temp,0b00000110		;Set entry mode:Increment RAM, dont shift display.
								;I/D=1(Increment RAM),S=0(no shift)
	rcall   SendCommand
;------ End of Init ----------
	ldi     temp,0b00001110 	;Display and cursor on, blink off.
								;D=1(Display on),C=1(display Cursor),B=0(no blink)
	rcall   SendCommand
	ldi     temp,0b00000001		;Display clear
	rcall   SendCommand
	rcall   delay2ms     		;Need to wait more than 1.6 msec after clear.
	rcall   delay2ms
	ret							

;********* Delay Routines for 3.6864 MHz Crystal ***********
Delay2us:	;Delay 2 us
	nop
	ret

Delay50us:	;Delay 50 us 
	push temp
	ldi	temp,$41		
d50l:
	dec	temp
	brne	d50l
	pop	temp
	ret		

Delay2ms:	;Delay 2 ms 
	push	temp
	ldi     temp,$0B
w2ms2:
	ldi temp2,$FA
w2ms1:
	dec     temp2
	brne    w2ms1
	dec  temp
	brne    w2ms2
	pop	temp
	ret


